'use strict';

const Promise = require('bluebird');
const util = require('util');
const cloneDeep = require('clone-deep');
const utils = require('punch-utils');

const ALEXA_URL = (index, categories) => `http://www.alexa.com/topsites/category;${index}/Top/${categories.join('/')}`;
const BATCH_SIZE = 10;



const ScraperManager = require('../index');
const scraperManager = new ScraperManager({
  scraperConfig: {
    strategy: {
      name: ScraperManager.STRATEGIES.HTTP
    },
    timeout: 8 * 1000,
    // Test to make sure result is an Alexa page and not a blocker.
    test: (result) => !!result.getHTML().match('Alexa - Top Sites by Category')
  },
  proxyManagerConfig: {
    proxyConfig: {
      // Logic for proxies to blacklist themselves.
      blacklistIf: (state) => {
        // Return if > 70% of the distinct URLs tested failed and at least
        // 5 have been attempted.
        return ((state.failed.distinctUrlCount/state.tried.distinctUrlCount) > .7) &&
            state.tried.distinctUrlCount > 5;
      }
    }
  },
  maxTry: 100
});

scraperManager.start();


/**
 *
 */
function scrape(categories, config) {
  !Array.isArray(categories) && (categories = [categories]);

  let urlCategories = categories.map(formatCategoryForUrl);

  // Do initial scrape to get pagination count and categories.
  return scraperManager.scrape([ALEXA_URL(0, urlCategories)])
  .then((results) => {
    // Get the subcategories and maxPage from the initial scrape.
    const response = formatResult(results[0]);

    // Generate all of the numbered urls based off of the maxPage.
    let urls = generateUrls(urlCategories, response.maxPage);

    return scrapeBatch(urls)
    .then(results => {
      let index = response.listings.indexed;
      results.forEach(result => {
        result.listings.data.forEach(listing => {
          // Check index of existing listings to make sure link hasn't already been
          // added (this is just a sanity precaution)
          if (index[listing.url]) return;
          response.listings.data.push(listing);
        });
        // Update the index of all processed urls.
        index = Object.assign(index, result.listings.indexed);
      });

      response.listings.data.sort((a, b) => a.ranking - b.ranking);
      return response;
    });
  });
}


/**
 *
 */
function scrapeBatch(urls, starting) {
  starting = starting || 0;
  let currentBatch = urls.slice(starting, starting + BATCH_SIZE);
  if (!currentBatch.length) return Promise.resolve([]);
  
  return scraperManager.scrape(currentBatch)
  .then(results => {
    results = results.map(result => formatResult(result, true));
    return scrapeBatch(urls, starting + BATCH_SIZE)
    .then(next => results.concat(next));
  });
}


/**
 *
 */
function formatResult(result, onlyListings) {
  if (!result) {
    // This shouldn't happen. Look into cause, remove this code when identified.
    console.log('Something went wrong.', result);
    return {listings: [], maxPage: 0, subcategories: []};
  }
  const $ = result.html.getParsed();

  const formatted = {listings: getListings($)};

  if (onlyListings) return formatted;
  return Object.assign(formatted, {
    maxPage: getMaxPage($),
    subcategories: getSubcategories($)
  });
}


/**
 *
 */
function getMaxPage($) {
  const $paginationTiles = $('.pagination-page');
  let maxPage = 0;
  $paginationTiles.each(function() {
    let page = Number($(this).text().trim());
    if (isNaN(page)) return;
    if (page > maxPage) maxPage = page;
  });
  return maxPage;
}


/**
 *
 */
function getSubcategories($) {
  const $subcategories = $('ul.subcategories a');
  const subcategories = [];

  $subcategories.each(function() {
    subcategories.push($(this).text().trim());
  });

  return subcategories;
};


/**
 *
 */
function getListings($) {
  const $listings = $('.site-listing');
  const listings = [];
  const indexed = {};

  $listings.each(function() {
    const formatted = {};
    const $listing = $(this);
    const $link = $listing.find('.desc-paragraph a');
    formatted.ranking = Number($listing.find('.count').text().trim());
    formatted.url = $link.text().trim().toLowerCase();
    formatted.alexa_url = `http://www.alexa.com${$link.attr('href').trim()}`;
    const $description = $listing.find('.description');
    formatted.description = $description.text().replace('…More', '').trim();
    listings.push(formatted);
    indexed[formatted.url] = formatted;
  });

  return {
    indexed,
    data: listings
  };
}


/**
 *
 */
function formatCategoryForUrl(category) {
  return category.replace(/\s/g, '_');
}


/**
 *
 */
function generateUrls(categories, cap) {
  cap = (cap + 1);
  // Handle weird case with alexa where you can get an extra page of links.
  if (cap == 21) cap = 22;
  let urls = [];
  // Not 0 indexed because first url has already been scraped.
  for (let i = 1; i < cap; i++) {
    urls.push(ALEXA_URL(i, categories));
  }
  return urls;
}


module.exports.scrape = scrape;
